# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/pdf/00_pdf_utils.ipynb.

# %% auto 0
__all__ = ['pdf_files', 'pdf_docs', 'pdf_to_docs', 'text_cat', 'docs_to_cats', 'add_cats_to_docs']

# %% ../../nbs/pdf/00_pdf_utils.ipynb 2
from dreamai.imports import *
from ..imports import *
from ..utils import *
from ..chains import *


# %% ../../nbs/pdf/00_pdf_utils.ipynb 4
def pdf_files(pdf_folder):
    return get_files(pdf_folder, extensions=[".pdf"])


def pdf_docs(pdf, splitter, proc=True):
    docs = PyPDFLoader(str(pdf)).load_and_split(text_splitter=splitter)
    if proc:
        docs = [proc_doc_text(doc) for doc in docs]
    return docs


def pdf_to_docs(
    path,
    splitter=None,
    chunk_size=200,
    chunk_overlap=60,
    separators=["\n\n", "\n", "(?<=\. )", " ", ""],
    add_start_index=True,
    proc=True,
):
    if splitter is None:
        splitter = RecursiveCharacterTextSplitter(
            chunk_size=chunk_size,
            chunk_overlap=chunk_overlap,
            separators=separators,
            add_start_index=add_start_index,
        )
    docs = []
    for pdf in resolve_data_path(path):
        if Path(pdf).suffix == ".pdf":
            d = pdf_docs(pdf, splitter=splitter, proc=proc)
            if len(d) > 0:
                docs.append(d)
            # docs.append(pdf_docs(pdf, splitter=splitter, proc=proc))
    return docs


def text_cat(
    text,
    cats_model,
    thresh=0.6,
    classes=["Work Experience", "Education", "Certifications", "Other"],
    other_class="Other",
):
    if other_class not in classes:
        classes.append(other_class)
    ot_id = classes.index(other_class)
    probs, ids = cats_model.predict_proba(text).detach().cpu().max(1)
    ids[probs < thresh] = ot_id
    pred_classes = np.array(classes)[ids]
    return pred_classes.tolist()


def docs_to_cats(docs, cats_model):
    return text_cat([doc.page_content for doc in docs], cats_model)

def add_cats_to_docs(docs, cats_model, key="category"):
    fn = partial(docs_to_cats, cats_model=cats_model)
    return add_docs_metadata(docs, fn, key)



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/ner/00_ner_utils.ipynb.

# %% auto 0
__all__ = ['EduNER', 'JobNER', 'load_edu_model', 'load_job_model', 'proc_ners', 'job_ner', 'edu_ner', 'work_ner', 'docs_to_ners',
           'add_ners_to_docs']

# %% ../../nbs/ner/00_ner_utils.ipynb 2
from ..imports import *
from ..chains import *
from ..utils import *
from ..pdf.utils import *
from pydantic import BaseModel, Field
from langchain.output_parsers import PydanticOutputParser

# %% ../../nbs/ner/00_ner_utils.ipynb 4
class EduNER(BaseModel):
    degree: str = Field(description="The degree of the education.", default=None)
    major: str = Field(description="The major of the education.", default=None)
    school: str = Field(description="The school of the education.", default=None)
    date: str = Field(description="The date of the education.", default=None)


class JobNER(BaseModel):
    role: str = Field(description="The role of the job.", default=None)
    company: str = Field(description="The company of the job.", default=None)
    duration: str = Field(description="The duration of the job.", default=None)

# %% ../../nbs/ner/00_ner_utils.ipynb 5
def load_edu_model(model_name="tner/deberta-v3-large-ontonotes5", device="cpu"):
    tokenizer = AutoTokenizer.from_pretrained(model_name)
    model = AutoModelForTokenClassification.from_pretrained(model_name)
    return pipeline(
        "ner",
        model=model,
        tokenizer=tokenizer,
        aggregation_strategy="simple",
        device=device,
    )


def load_job_model(model_name="ismail-lucifer011/autotrain-job_all-903929564", device="cpu"):
    tokenizer = AutoTokenizer.from_pretrained(model_name)
    model = AutoModelForTokenClassification.from_pretrained(model_name)
    return pipeline(
        "ner",
        model=model,
        tokenizer=tokenizer,
        aggregation_strategy="simple",
        device=device,
    )


def proc_ners(ners, ner_dict={"institute": "", "date": ""}, thresh=3):
    ner_dict2 = copy.deepcopy(ner_dict)
    org_key = "institute" if "institute" in ner_dict2 else "company"
    mapper = {
        "ORG": org_key,
        "FAC": org_key,
        "GPE": org_key,
        "LOC": org_key,
        "Job": "role",
        "WORK_OF_ART": "degree",
        "DATE": "date",
    }
    ner_dicts = []
    for ner in ners:
        if len(ner) == 0:
            ner_dicts.append({})
            continue
        try:
            for d in ner:
                eg = d["entity_group"]
                w = " " + d["word"].strip()
                k = mapper.get(eg, None)
                if k is not None and ner_dict2.get(k, None) is not None and not w.startswith("##"):
                    ner_dict2[k] = (ner_dict2[k] + w).strip()
            res = {k: v for k, v in ner_dict2.items() if len(v) > thresh}
            if res.get(org_key, None) is not None:
                ner_dicts.append(res)
            else:
                ner_dicts.append({})
            ner_dict2 = copy.deepcopy(ner_dict)
        except Exception as e:
            msg.fail(f"proc_ners failed with error: {e}", spaced=True)
            ner_dicts.append({})
            ner_dict2 = copy.deepcopy(ner_dict)
    return ner_dicts


def job_ner(docs, e_ner, j_ner):
    return j_ner(docs), e_ner(docs)


def edu_ner(docs, e_ner, ner_dict={"institute": "", "date": ""}):
    ners = e_ner(docs)
    return proc_ners(ners, ner_dict)


def work_ner(docs, e_ner, j_ner, ner_dict={"company": "", "date": ""}):
    ner1, ner2 = job_ner(docs, e_ner, j_ner)
    ners = [n1 + n2 for n1, n2 in zip(ner1, ner2)]
    return proc_ners(ners, ner_dict)


def docs_to_ners(docs, e_ner, j_ner):
    ners = [{}] * len(docs)
    work_docs = np.array(
        [
            [i, doc.page_content]
            for i, doc in enumerate(docs)
            if doc.metadata.get("category", None) == "Work Experience"
        ]
    )
    work_docs_idx = work_docs[:, 0].astype(int)
    work_docs = work_docs[:, 1].tolist()
    try:
        work_ners = work_ner(work_docs, e_ner, j_ner)
    except Exception as e:
        msg.fail(f"work_ner failed with error: {e}", spaced=True)
        work_ners = [{}] * len(work_docs)
    for i, doc in enumerate(work_docs_idx):
        ners[doc] = work_ners[i]
    edu_docs = np.array(
        [
            [i, doc.page_content]
            for i, doc in enumerate(docs)
            if doc.metadata.get("category", None) == "Education"
        ]
    )
    edu_docs_idx = edu_docs[:, 0].astype(int)
    edu_docs = edu_docs[:, 1].tolist()
    try:
        edu_ners = edu_ner(edu_docs, e_ner)
    except Exception as e:
        msg.fail(f"edu_ner failed with error: {e}", spaced=True)
        edu_ners = [{}] * len(edu_docs)
    for i, doc in enumerate(edu_docs_idx):
        ners[doc] = edu_ners[i]
    return ners

def add_ners_to_docs(docs, e_ner, j_ner, key="ner"):
    fn = partial(docs_to_ners, e_ner=e_ner, j_ner=j_ner)
    return add_docs_metadata(docs, fn, key)



# Autogenerated by nbdev

d = { 'settings': { 'branch': 'main',
                'doc_baseurl': '/langchain_ray',
                'doc_host': 'https://HamzaFarhan.github.io',
                'git_url': 'https://github.com/HamzaFarhan/langchain_ray',
                'lib_path': 'langchain_ray'},
  'syms': { 'langchain_ray.api.app': {},
            'langchain_ray.api.app_deployment': {},
            'langchain_ray.api.imports': {},
            'langchain_ray.chains': { 'langchain_ray.chains.add_str_to_docs_chain': ( 'chains.html#add_str_to_docs_chain',
                                                                                      'langchain_ray/chains.py'),
                                      'langchain_ray.chains.chain_fn': ('chains.html#chain_fn', 'langchain_ray/chains.py'),
                                      'langchain_ray.chains.chain_fn_args': ('chains.html#chain_fn_args', 'langchain_ray/chains.py'),
                                      'langchain_ray.chains.chainfn_input': ('chains.html#chainfn_input', 'langchain_ray/chains.py'),
                                      'langchain_ray.chains.chainfn_output': ('chains.html#chainfn_output', 'langchain_ray/chains.py'),
                                      'langchain_ray.chains.docs_to_json_chain': ( 'chains.html#docs_to_json_chain',
                                                                                   'langchain_ray/chains.py'),
                                      'langchain_ray.chains.noop_chain': ('chains.html#noop_chain', 'langchain_ray/chains.py'),
                                      'langchain_ray.chains.ray_chain': ('chains.html#ray_chain', 'langchain_ray/chains.py'),
                                      'langchain_ray.chains.ray_chain_fn': ('chains.html#ray_chain_fn', 'langchain_ray/chains.py'),
                                      'langchain_ray.chains.transform_chain': ('chains.html#transform_chain', 'langchain_ray/chains.py')},
            'langchain_ray.driver.ingress_driver': {},
            'langchain_ray.driver.redis_kv_store': {},
            'langchain_ray.imports': {},
            'langchain_ray.indexing.chains': { 'langchain_ray.indexing.chains.add_ems_to_docs_chain': ( 'indexing/indexing_chains.html#add_ems_to_docs_chain',
                                                                                                        'langchain_ray/indexing/chains.py'),
                                               'langchain_ray.indexing.chains.docs_to_faiss_chain': ( 'indexing/indexing_chains.html#docs_to_faiss_chain',
                                                                                                      'langchain_ray/indexing/chains.py'),
                                               'langchain_ray.indexing.chains.index_query_chain': ( 'indexing/indexing_chains.html#index_query_chain',
                                                                                                    'langchain_ray/indexing/chains.py'),
                                               'langchain_ray.indexing.chains.pdf_to_faiss_chain': ( 'indexing/indexing_chains.html#pdf_to_faiss_chain',
                                                                                                     'langchain_ray/indexing/chains.py')},
            'langchain_ray.indexing.utils': { 'langchain_ray.indexing.utils.add_ems_to_docs': ( 'indexing/indexing_utils.html#add_ems_to_docs',
                                                                                                'langchain_ray/indexing/utils.py'),
                                              'langchain_ray.indexing.utils.docs_to_ems': ( 'indexing/indexing_utils.html#docs_to_ems',
                                                                                            'langchain_ray/indexing/utils.py'),
                                              'langchain_ray.indexing.utils.docs_to_faiss': ( 'indexing/indexing_utils.html#docs_to_faiss',
                                                                                              'langchain_ray/indexing/utils.py'),
                                              'langchain_ray.indexing.utils.index_names': ( 'indexing/indexing_utils.html#index_names',
                                                                                            'langchain_ray/indexing/utils.py'),
                                              'langchain_ray.indexing.utils.search_faiss': ( 'indexing/indexing_utils.html#search_faiss',
                                                                                             'langchain_ray/indexing/utils.py')},
            'langchain_ray.ner.chains': { 'langchain_ray.ner.chains.add_ners_to_docs_chain': ( 'ner/ner_chains.html#add_ners_to_docs_chain',
                                                                                               'langchain_ray/ner/chains.py')},
            'langchain_ray.ner.utils': { 'langchain_ray.ner.utils.EduNER': ('ner/ner_utils.html#eduner', 'langchain_ray/ner/utils.py'),
                                         'langchain_ray.ner.utils.JobNER': ('ner/ner_utils.html#jobner', 'langchain_ray/ner/utils.py'),
                                         'langchain_ray.ner.utils.add_ners_to_docs': ( 'ner/ner_utils.html#add_ners_to_docs',
                                                                                       'langchain_ray/ner/utils.py'),
                                         'langchain_ray.ner.utils.docs_to_ners': ( 'ner/ner_utils.html#docs_to_ners',
                                                                                   'langchain_ray/ner/utils.py'),
                                         'langchain_ray.ner.utils.edu_ner': ('ner/ner_utils.html#edu_ner', 'langchain_ray/ner/utils.py'),
                                         'langchain_ray.ner.utils.job_ner': ('ner/ner_utils.html#job_ner', 'langchain_ray/ner/utils.py'),
                                         'langchain_ray.ner.utils.load_edu_model': ( 'ner/ner_utils.html#load_edu_model',
                                                                                     'langchain_ray/ner/utils.py'),
                                         'langchain_ray.ner.utils.load_job_model': ( 'ner/ner_utils.html#load_job_model',
                                                                                     'langchain_ray/ner/utils.py'),
                                         'langchain_ray.ner.utils.proc_ners': ( 'ner/ner_utils.html#proc_ners',
                                                                                'langchain_ray/ner/utils.py'),
                                         'langchain_ray.ner.utils.work_ner': ('ner/ner_utils.html#work_ner', 'langchain_ray/ner/utils.py')},
            'langchain_ray.pdf.chains': { 'langchain_ray.pdf.chains.add_cats_to_docs_chain': ( 'pdf/pdf_chains.html#add_cats_to_docs_chain',
                                                                                               'langchain_ray/pdf/chains.py'),
                                          'langchain_ray.pdf.chains.pdf_to_docs_chain': ( 'pdf/pdf_chains.html#pdf_to_docs_chain',
                                                                                          'langchain_ray/pdf/chains.py')},
            'langchain_ray.pdf.ner': { 'langchain_ray.pdf.ner.add_ners_to_docs': ( 'pdf/pdf_ner.html#add_ners_to_docs',
                                                                                   'langchain_ray/pdf/ner.py'),
                                       'langchain_ray.pdf.ner.docs_to_ners': ('pdf/pdf_ner.html#docs_to_ners', 'langchain_ray/pdf/ner.py'),
                                       'langchain_ray.pdf.ner.edu_ner': ('pdf/pdf_ner.html#edu_ner', 'langchain_ray/pdf/ner.py'),
                                       'langchain_ray.pdf.ner.job_ner': ('pdf/pdf_ner.html#job_ner', 'langchain_ray/pdf/ner.py'),
                                       'langchain_ray.pdf.ner.load_job_model': ( 'pdf/pdf_ner.html#load_job_model',
                                                                                 'langchain_ray/pdf/ner.py'),
                                       'langchain_ray.pdf.ner.load_ner_model': ( 'pdf/pdf_ner.html#load_ner_model',
                                                                                 'langchain_ray/pdf/ner.py'),
                                       'langchain_ray.pdf.ner.proc_ners': ('pdf/pdf_ner.html#proc_ners', 'langchain_ray/pdf/ner.py'),
                                       'langchain_ray.pdf.ner.work_ner': ('pdf/pdf_ner.html#work_ner', 'langchain_ray/pdf/ner.py')},
            'langchain_ray.pdf.utils': { 'langchain_ray.pdf.utils.add_cats_to_docs': ( 'pdf/pdf_utils.html#add_cats_to_docs',
                                                                                       'langchain_ray/pdf/utils.py'),
                                         'langchain_ray.pdf.utils.docs_to_cats': ( 'pdf/pdf_utils.html#docs_to_cats',
                                                                                   'langchain_ray/pdf/utils.py'),
                                         'langchain_ray.pdf.utils.pdf_docs': ('pdf/pdf_utils.html#pdf_docs', 'langchain_ray/pdf/utils.py'),
                                         'langchain_ray.pdf.utils.pdf_files': ( 'pdf/pdf_utils.html#pdf_files',
                                                                                'langchain_ray/pdf/utils.py'),
                                         'langchain_ray.pdf.utils.pdf_to_docs': ( 'pdf/pdf_utils.html#pdf_to_docs',
                                                                                  'langchain_ray/pdf/utils.py'),
                                         'langchain_ray.pdf.utils.text_cat': ('pdf/pdf_utils.html#text_cat', 'langchain_ray/pdf/utils.py')},
            'langchain_ray.remote_utils': { 'langchain_ray.remote_utils.bucket_count': ( 'remote_utils.html#bucket_count',
                                                                                         'langchain_ray/remote_utils.py'),
                                            'langchain_ray.remote_utils.bucket_del': ( 'remote_utils.html#bucket_del',
                                                                                       'langchain_ray/remote_utils.py'),
                                            'langchain_ray.remote_utils.bucket_dl': ( 'remote_utils.html#bucket_dl',
                                                                                      'langchain_ray/remote_utils.py'),
                                            'langchain_ray.remote_utils.bucket_move': ( 'remote_utils.html#bucket_move',
                                                                                        'langchain_ray/remote_utils.py'),
                                            'langchain_ray.remote_utils.bucket_up': ( 'remote_utils.html#bucket_up',
                                                                                      'langchain_ray/remote_utils.py'),
                                            'langchain_ray.remote_utils.get_local_path': ( 'remote_utils.html#get_local_path',
                                                                                           'langchain_ray/remote_utils.py'),
                                            'langchain_ray.remote_utils.gsutil_bucket': ( 'remote_utils.html#gsutil_bucket',
                                                                                          'langchain_ray/remote_utils.py'),
                                            'langchain_ray.remote_utils.gsutil_src': ( 'remote_utils.html#gsutil_src',
                                                                                       'langchain_ray/remote_utils.py'),
                                            'langchain_ray.remote_utils.handle_input_path': ( 'remote_utils.html#handle_input_path',
                                                                                              'langchain_ray/remote_utils.py'),
                                            'langchain_ray.remote_utils.is_bucket': ( 'remote_utils.html#is_bucket',
                                                                                      'langchain_ray/remote_utils.py')},
            'langchain_ray.utils': { 'langchain_ray.utils.add_docs_metadata': ('utils.html#add_docs_metadata', 'langchain_ray/utils.py'),
                                     'langchain_ray.utils.add_str_to_docs': ('utils.html#add_str_to_docs', 'langchain_ray/utils.py'),
                                     'langchain_ray.utils.bold_text': ('utils.html#bold_text', 'langchain_ray/utils.py'),
                                     'langchain_ray.utils.cid_to_char': ('utils.html#cid_to_char', 'langchain_ray/utils.py'),
                                     'langchain_ray.utils.docs_to_json': ('utils.html#docs_to_json', 'langchain_ray/utils.py'),
                                     'langchain_ray.utils.gen_random_string': ('utils.html#gen_random_string', 'langchain_ray/utils.py'),
                                     'langchain_ray.utils.is_doc': ('utils.html#is_doc', 'langchain_ray/utils.py'),
                                     'langchain_ray.utils.list_or_array': ('utils.html#list_or_array', 'langchain_ray/utils.py'),
                                     'langchain_ray.utils.nested_list': ('utils.html#nested_list', 'langchain_ray/utils.py'),
                                     'langchain_ray.utils.print_doc': ('utils.html#print_doc', 'langchain_ray/utils.py'),
                                     'langchain_ray.utils.proc_doc_text': ('utils.html#proc_doc_text', 'langchain_ray/utils.py'),
                                     'langchain_ray.utils.process_text': ('utils.html#process_text', 'langchain_ray/utils.py'),
                                     'langchain_ray.utils.unnest_list': ('utils.html#unnest_list', 'langchain_ray/utils.py')}}}



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/02_remote_utils.ipynb.

# %% auto 0
__all__ = ['is_bucket', 'gsutil_bucket', 'gsutil_src', 'bucket_del', 'bucket_count', 'bucket_move', 'bucket_up', 'bucket_dl',
           'get_local_path', 'handle_input_path']

# %% ../nbs/02_remote_utils.ipynb 2
from dreamai.imports import *
from .imports import *
from .utils import *


# %% ../nbs/02_remote_utils.ipynb 4
def is_bucket(p):
    return str(p).startswith("gs://")


def gsutil_bucket(bucket):
    if not str(bucket).startswith("gs://"):
        bucket = "gs://" + str(bucket)
    if bucket[-1] != "/" and Path(bucket).suffix != ".json":
        bucket += "/"
    return bucket


def gsutil_src(folder):
    if Path(folder).suffix != "":
        return str(folder)
    folder = str(folder)
    if folder[-1] != "/":
        folder += "/"
    folder += "*"
    return folder


def bucket_del(bucket):
    msg.info(f"Deleting {bucket}.", spaced=True)
    gu = shutil.which("gsutil")
    bucket = gsutil_bucket(bucket)
    subprocess.run([gu, "-m", "rm", "-r", bucket])


def bucket_count(bucket):
    gu = shutil.which("gsutil")
    bucket = gsutil_src(bucket)
    cmd = [gu, "ls", "-l", bucket]
    res = subprocess.run(cmd, stdout=subprocess.PIPE)
    res = res.stdout.decode("utf-8").split("\n")
    return len(res) - 2


def bucket_move(folder, bucket):
    msg.info(f"Moving {folder} to {bucket}.", spaced=True)
    gu = shutil.which("gsutil")
    bucket = gsutil_bucket(bucket)
    folder = gsutil_src(folder)
    files = get_files(folder, make_str=True)
    if len(files) == 1:
        folder = files[0]
    subprocess.run([gu, "-m", "mv", folder, bucket])


def bucket_up(folder, bucket, only_new=True):
    msg.info(f"Uploading {folder} to {bucket}.", spaced=True)
    gu = shutil.which("gsutil")
    bucket = gsutil_bucket(bucket)
    folder = gsutil_src(folder)
    cmd = [gu, "-m", "cp"]
    if only_new:
        cmd.append("-n")
    files = get_files(folder, make_str=True)
    if len(files) == 1:
        folder = files[0]
    else:
        cmd.append("-r")
    subprocess.run(cmd + [folder, bucket])


def bucket_dl(bucket, folder, only_new=True):
    msg.info(f"Downloading {bucket} to {folder}.", spaced=True)
    gu = shutil.which("gsutil")
    bucket = gsutil_bucket(bucket)
    bucket = gsutil_src(bucket)
    os.makedirs(folder, exist_ok=True)
    cmd = [gu, "-m", "cp", "-r"]
    if only_new:
        cmd.append("-n")
    subprocess.run(cmd + [bucket, folder])


def get_local_path(remote_path, local_path):
    if is_bucket(remote_path):
        if Path(remote_path).suffix != "":
            return Path(local_path), remote_path
        return Path(local_path) / Path(remote_path).name, remote_path
    else:
        return Path(remote_path), remote_path


def handle_input_path(path, local_path=None, task_id=gen_random_string(16), only_new=True):
    path = str(path)
    if not is_bucket(path):
        return path, path
    if local_path is None:
        local_path = Path(f"/tmp/{task_id}")
    os.makedirs(local_path, exist_ok=True)
    local_path, bucket = get_local_path(path, local_path)
    bucket_dl(bucket, local_path, only_new=only_new)
    return local_path, bucket



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/ner/01_ner_chains.ipynb.

# %% auto 0
__all__ = ['add_ners_to_docs_chain']

# %% ../../nbs/ner/01_ner_chains.ipynb 2
from ..imports import *
from ..chains import *
from ..utils import *
from ..pdf.utils import *
from ..pdf.chains import *
from .utils import *


# %% ../../nbs/ner/01_ner_chains.ipynb 4
def add_ners_to_docs_chain(
    e_ner,
    j_ner,
    input_variables=["docs"],
    output_variables=["ner_docs"],
    verbose=False,
):
    "Chain that adds the NERs to a list of `Documents` usung `e_ner` and `j_ner`."
    return transform_chain(
        add_ners_to_docs,
        transform_kwargs=dict(e_ner=e_ner, j_ner=j_ner),
        input_variables=input_variables,
        output_variables=output_variables,
        vars_kwargs_mapping={input_variables[0]: "docs"},
        verbose=verbose,
    )



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/01_chains.ipynb.

# %% auto 0
__all__ = ['chainfn_input', 'chainfn_output', 'chain_fn_args', 'chain_fn', 'transform_chain', 'ray_chain_fn', 'ray_chain',
           'noop_chain', 'docs_to_json_chain', 'add_str_to_docs_chain']

# %% ../nbs/01_chains.ipynb 2
from dreamai.imports import *
from .imports import *
from .utils import *

# %% ../nbs/01_chains.ipynb 4
def chainfn_input(data, tfm, tfm_kwargs={}, input_variables=["df"]):
    try:
        for k in input_variables:
            tfm_kwargs[k] = data.get(k, None)
        fn_args = inspect.signature(tfm).parameters
        pos_args = [k for k in fn_args if fn_args[k].default == inspect._empty]
        for pk, ik in zip(pos_args, input_variables):
            if pk != "kwargs":
                tfm_kwargs[pk] = tfm_kwargs.pop(ik)
        if "kwargs" not in fn_args:
            tfm_kwargs = {k: v for k, v in tfm_kwargs.items() if k in fn_args.keys()}
    except Exception as e:
        msg.fail(f"Error in chainfn_input: {e}", spaced=True)
    return tfm_kwargs


# def chain_fn(data, tfm, tfm_kwargs={}, input_variables=["df"], output_variables=["df"]):
#     tfm_kwargs = chainfn_input(data, tfm, tfm_kwargs, input_variables)
#     # print(f'\n\nTFM: {tfm}\n\n')
#     # print(f'\n\nTFM KWARGS: {tfm_kwargs.keys()}\n\n')
#     fn_res = tfm(**tfm_kwargs)
#     return chainfn_output(fn_res, output_variables)


def chainfn_output(fn_res, output_variables=["df"]):
    if not list_or_tuple(fn_res):
        fn_res = [fn_res]
    if len(fn_res) > len(output_variables):
        fn_res = [fn_res]
    return {k: r for k, r in zip(output_variables, fn_res)}


def chain_fn_args(data, tfm, tfm_kwargs={}, vars_kwargs_mapping={}):
    try:
        tfm_kwargs = {**data, **tfm_kwargs}
        for k, v in vars_kwargs_mapping.items():
            tfm_kwargs[v] = tfm_kwargs.pop(k)
        fn_args = inspect.signature(tfm).parameters
        if "kwargs" not in fn_args:
            tfm_kwargs = {k: v for k, v in tfm_kwargs.items() if k in fn_args.keys()}
    except Exception as e:
        msg.fail(f"Error in chain_fn_args: {e}", spaced=True)
    return tfm_kwargs


def chain_fn(data, tfm, tfm_kwargs={}, vars_kwargs_mapping={}, output_variables=["df"]):
    tfm_kwargs = chain_fn_args(data, tfm, tfm_kwargs, vars_kwargs_mapping)
    # print(f"\n\nTFM: {tfm}\n\n")
    # print(f"\n\nTFM KWARGS: {tfm_kwargs.keys()}\n\n")
    fn_res = tfm(**tfm_kwargs)
    fn_res = {**data, **chainfn_output(fn_res, output_variables)}
    # print(f"\n\nFN RES: {fn_res}\n\n")
    return fn_res


def transform_chain(
    transform,
    transform_kwargs={},
    vars_kwargs_mapping={},  # For mapping `input_variables` to `transform_kwargs`
    input_variables=["df"],
    output_variables=["df"],
    verbose=False,
):
    "Return a `transform` function wrapped in a `TransformChain`."
    return TransformChain(
        input_variables=input_variables,
        output_variables=output_variables,
        transform=partial(
            chain_fn,
            tfm=transform,
            tfm_kwargs=transform_kwargs,
            vars_kwargs_mapping=vars_kwargs_mapping,
            output_variables=output_variables,
        ),
        verbose=verbose,
    )


def ray_chain_fn(data, chain, block_size=1500, num_cpus=8, num_gpus=1):
    """
    This function is used to run a given chain function on a Ray cluster.
    It splits the input data into blocks, and runs the chain function on each block in parallel.
    The results are then combined and returned.
    """

    try:
        # If the block size is None, run the chain function on the entire dataset.
        if block_size is None:
            return chain(data)

        # Convert the data into a dictionary of lists.
        ray_data = {k: v for k, v in data.items()}
        for k, v in ray_data.items():
            if not is_list(v):
                ray_data[k] = [v]

        # Get the maximum length of any list in the data dictionary.
        max_len = max([len(v) for v in ray_data.values()])

        # If the maximum length is 1 or less than the block size, run the chain function on the entire dataset.
        if max_len == 1 or max_len <= block_size:
            return chain(data)

        # If the maximum length is greater than the block size, split the data into blocks.
        for k, v in ray_data.items():
            if len(v) < max_len:
                ray_data[k] = v * max_len

        # Convert the data dictionary into a Pandas DataFrame.
        df = pd.DataFrame(ray_data)

        # Initialize Ray.
        ray.init(ignore_reinit_error=True)

        # Calculate the number of blocks to split the data into.
        num_blocks = int(np.ceil(len(df) / block_size))
        msg.info(f"Running chain on {num_blocks} blocks.", spaced=True)

        # Calculate the number of CPUs and GPUs to use per block.
        num_cpus = min(ray.available_resources()["CPU"], num_cpus)
        num_cpus /= num_blocks
        if num_gpus is not None:
            num_gpus = min(ray.available_resources().get("GPU", 0), num_gpus)
            if num_gpus > 0:
                num_gpus /= num_blocks
                num_cpus = None
            else:
                num_gpus = None
        # Convert the Pandas DataFrame into a Ray dataset and run the chain function on each block.
        ds = rd.from_pandas(df).repartition(num_blocks)
        res = (
            ds.map_batches(
                lambda x: chain(x),
                batch_size=block_size,
                num_cpus=num_cpus,
                num_gpus=num_gpus,
            )
            .to_pandas()
            .to_dict(orient="list")
        )

        return res
    except Exception as e:
        msg.fail(f"Error in ray_chain_fn: {e}", spaced=True)
        return chain(data)


def ray_chain(chain, block_size=1500, num_cpus=8, num_gpus=1, verbose=False):
    """
    This function is used to run a given chain function on a Ray cluster if necessary.
    """
    tfm = partial(
        ray_chain_fn,
        chain=chain,
        block_size=block_size,
        num_cpus=num_cpus,
        num_gpus=num_gpus,
    )
    input_variables = chain.input_keys
    output_variables = chain.output_keys
    return TransformChain(
        input_variables=input_variables,
        output_variables=output_variables,
        transform=tfm,
        verbose=verbose,
    )


def noop_chain():
    return transform_chain(noop)


def docs_to_json_chain(
    json_folder,
    data={},
    data_key="data",
    with_metadata=True,
    with_content=True,
    indent=None,
    input_variables=["docs"],
    output_variables=["docs"],
    verbose=False,
):
    return transform_chain(
        docs_to_json,
        transform_kwargs=dict(
            json_folder=json_folder,
            data=data,
            data_key=data_key,
            with_metadata=with_metadata,
            with_content=with_content,
            indent=indent,
        ),
        vars_kwargs_mapping={input_variables[0]: "docs"},
        input_variables=input_variables,
        output_variables=output_variables,
        verbose=verbose,
    )
    
def add_str_to_docs_chain(
    str,
    input_variables=["docs"],
    output_variables=["cat_docs"],
    verbose=False,
):
    "Chain that adds a string to the docs"
    return transform_chain(
        add_str_to_docs,
        transform_kwargs=dict(str=str),
        input_variables=input_variables,
        output_variables=output_variables,
        vars_kwargs_mapping={input_variables[0]: "docs"},
        verbose=verbose,
    )



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/indexing/01_indexing_chains.ipynb.

# %% auto 0
__all__ = ['add_ems_to_docs_chain', 'docs_to_faiss_chain', 'pdf_to_faiss_chain', 'index_query_chain']

# %% ../../nbs/indexing/01_indexing_chains.ipynb 2
from ..imports import *
from ..chains import *
from ..utils import *
from .utils import *
from ..pdf.chains import *


# %% ../../nbs/indexing/01_indexing_chains.ipynb 4
def add_ems_to_docs_chain(
    ems_model,
    input_variables=["docs"],
    output_variables=["ems_docs"],
    verbose=False,
):
    "Chain that takes a list of `Documents` and adds `embeddings` to each `Document`."
    return transform_chain(
        add_ems_to_docs,
        transform_kwargs=dict(ems_model=ems_model),
        input_variables=input_variables,
        output_variables=output_variables,
        vars_kwargs_mapping={input_variables[0]: "docs"},
        verbose=verbose,
    )


def docs_to_faiss_chain(
    ems_model,
    index_folder,
    index_name,
    input_variables=["docs"],
    output_variables=["docs"],
    verbose=False,
):
    "Chain that takes a list of `Documents` and adds them to a `FAISS` index in `index_folder`."
    return transform_chain(
        docs_to_faiss,
        input_variables=input_variables,
        output_variables=output_variables,
        transform_kwargs={
            "ems_model": ems_model,
            "index_folder": index_folder,
            "index_name": index_name,
        },
        vars_kwargs_mapping={input_variables[0]: "docs"},
        verbose=verbose,
    )


def pdf_to_faiss_chain(
    ems_model,  # The model to use for vectorestore embeddings.
    index_folder,  # The folder to store the FAISS index.
    index_name,  # The name of the FAISS index.
    input_variables=["pdf_path"],
    output_variables=["docs"],
    chunk_size=200,  # The number of characters per Document.
    chunk_overlap=20,  # The number of characters to overlap between Documents.
    block_size=1500,  # The number of PDFs to process in a single Ray task.
    num_cpus=12,  # The number of CPUs to use for Ray.
    num_gpus=1,  # The number of GPUs to use for Ray.
    verbose=False,
):
    """
    Chain that adds PDFs to `FAISS` indexes in `index_folder`.
    If there are more than `block_size` PDFs, indexing will be distributed using `Ray`.
    """
    docs_chain = pdf_to_docs_chain(
        chunk_size=chunk_size,
        chunk_overlap=chunk_overlap,
        input_variables=input_variables,
        output_variables=["dc"],
        verbose=verbose,
    )
    faiss_chain = docs_to_faiss_chain(
        ems_model=ems_model,
        index_folder=index_folder,
        index_name=index_name,
        input_variables=["dc"],
        output_variables=output_variables,
        verbose=verbose,
    )
    faiss_chain = ray_chain(
        faiss_chain,
        block_size=block_size,
        num_cpus=num_cpus,
        num_gpus=num_gpus,
        verbose=verbose,
    )
    return SequentialChain(
        chains=[docs_chain, faiss_chain],
        input_variables=input_variables,
        output_variables=output_variables,
        verbose=verbose,
    )


def index_query_chain(
    ems_model,  # The SentenceTransformer model to use for vectorestore embeddings.
    index_folder,  # The folder with the FAISS indexes.
    index_name,  # The name of the FAISS index.
    input_variables=["query", "k"],
    output_variables=["search_results"],
    block_size=10,  # The number of indexes to process in a single Ray task.
    num_cpus=12,  # The number of CPUs to use for Ray.
    num_gpus=1,  # The number of GPUs to use for Ray.
    verbose=False,
):
    """
    Chain that takes a `query` and returns the top `k` results from the `FAISS` indexes in `index_folder`.
    If there are more than `block_size` indexes, search will be distributed using `Ray`.
    """

    index_names_chain = transform_chain(
        index_names,
        transform_kwargs={"index_folder": index_folder, "index_name": index_name},
        input_variables=["k"],
        output_variables=["index_names"],
    )

    search_faiss_chain = transform_chain(
        search_faiss,
        transform_kwargs={"index_folder": index_folder, "ems_model": ems_model},
        input_variables=["index_names", "query", "k"],
        output_variables=["res"],
    )

    search_faiss_chain = ray_chain(
        search_faiss_chain, block_size=block_size, num_cpus=num_cpus, num_gpus=num_gpus
    )

    def flatten_res(res, k):
        if is_list(k):
            k = k[0]
        return [sorted(flatten_list(res), key=lambda x: x[1])[:k]]

    res_chain = transform_chain(
        flatten_res,
        input_variables=["res", "k"],
        output_variables=output_variables,
    )

    return SequentialChain(
        chains=[index_names_chain, search_faiss_chain, res_chain],
        input_variables=input_variables,
        output_variables=output_variables,
        verbose=verbose,
    )




# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/00_utils.ipynb.

# %% auto 0
__all__ = ['is_doc', 'gen_random_string', 'list_or_array', 'nested_list', 'unnest_list', 'cid_to_char', 'process_text',
           'proc_doc_text', 'bold_text', 'print_doc', 'docs_to_json', 'add_docs_metadata', 'add_str_to_docs']

# %% ../nbs/00_utils.ipynb 2
from dreamai.imports import *
from .imports import *


# %% ../nbs/00_utils.ipynb 4
def is_doc(x):
    return isinstance(x, Document)


def gen_random_string(length):
    return str(uuid.uuid4()).replace("-", "")[:length]


def list_or_array(x):
    return is_list(x) or is_array(x)


def nested_list(x):
    is_nested = list_or_array(x) and list_or_array(x[0])
    if not list_or_array(x):
        x = [x]
    if not list_or_array(x[0]):
        x = [x]
    return x, is_nested


def unnest_list(x):
    if is_list(x) and is_list(x[0]) and len(x) == 1:
        x = x[0]
    return x


def cid_to_char(cidx: str):
    try:
        return chr(int(re.findall(r"\(cid\:(\d+)\)", cidx)[0]) + 29)
    except:
        return cidx


def process_text(text: str):
    text = text.strip()
    text = demoji.replace(text, "")
    text = clean(
        text,
        # no_urls=True,
        no_emails=True,
        no_phone_numbers=True,
        no_currency_symbols=True,
        # replace_with_url="",
        replace_with_email="",
        replace_with_phone_number="",
        replace_with_currency_symbol="",
    )
    text = cid_to_char(text)
    text = re.sub("\xa0", " ", text)
    text = re.sub(r"\uf0b7", " ", text)
    text = re.sub(r"\(cid:\d{0,3}\)", " ", text)
    text = re.sub(r"•", "", text)
    text = re.sub(r"●", "", text)
    text = re.sub(r"▪", "", text)
    text = re.sub(r"", "", text)
    text = re.sub(r"➢", "", text)
    text = re.sub(r"\u2b9a", "", text)
    text = re.sub(r"\u201c", "", text)
    text = re.sub(r"\u201d", "", text)
    text = re.sub(r"\u2013", " ", text)
    text = re.sub(r"\u2019", "'", text)
    text = re.sub(r"\u2018", "'", text)
    text = re.sub(r"\u00f4", " ", text)
    text = re.sub(r"\u00f6", "o", text)
    text = re.sub(r"\u00e9", "e", text)
    text = re.sub(r"\u00e8", "e", text)
    text = re.sub(r"\u00e7", " ", text)
    text = re.sub(r"\u00a7", "", text)
    text = re.sub(r"\u00e3", "a", text)
    text = re.sub(r"\uf0a7", "", text)
    text = re.sub(r"\uf076", "", text)
    text = re.sub(r"\u00ad", "", text)
    text = re.sub(r"\u00ab", "", text)
    text = re.sub(r"\u00bb", "", text)
    text = re.sub(r"\uf02d", "", text)
    text = re.sub(r"\uf0fc", "", text)
    text = re.sub(r"\uf06e", "", text)
    text = re.sub(r"\uf07a", "", text)
    text = re.sub(r"\ufb01", "fi", text)
    text = re.sub(r"\ufb00", "ff", text)
    text = re.sub(r"\uf0d8", "", text)
    text = re.sub(r"\u00b7", "", text)
    text = re.sub("\t", " ", text)
    text = re.sub(" +", " ", text)
    return text.strip()


def proc_doc_text(doc):
    doc.page_content = process_text(doc.page_content)
    return doc


def bold_text(text):
    return "\033[1m" + text + "\033[0m"


def print_doc(doc):
    print(f"{bold_text('Page_Content:')} {doc.page_content}\n")
    print(f"{bold_text('Metadata:')} {doc.metadata}\n")


def docs_to_json(
    docs,
    json_folder,
    data={},
    data_key="data",
    with_metadata=True,
    with_content=False,
    indent=None,
):
    if is_list(data):
        data = {data_key: data}
    os.makedirs(json_folder, exist_ok=True)
    for i, doc in enumerate(flatten_list(docs)):
        doc_dict = {}
        if with_content:
            doc_dict["page_content"] = doc.page_content
        if with_metadata:
            doc_dict["metadata"] = doc.metadata
        for k, v in data.items():
            if is_list(v) and len(v) == len(docs):
                doc_dict[k] = v[i]
        if len(doc_dict) == 0:
            doc_dict = {"page_content": doc.page_content, "metadata": doc.metadata}
        source = Path(doc.metadata["source"])
        json_path = (Path(json_folder) / source.stem).with_suffix(".json")
        if json_path.exists():
            json_path = find_alternate_path(json_path, first_idx=1, verbose=False)
        with open(json_path, "w") as f:
            json.dump(doc_dict, f, indent=indent)
    return docs


def add_docs_metadata(docs, fn=None, key="new_meta"):
    if fn is None:
        return docs
    docs, is_nested = nested_list(docs)
    for docs_ in docs:
        fn_res = fn(docs_)
        for doc, res in zip(docs_, fn_res):
            doc.metadata[key] = res
    return docs if is_nested else docs[0]

def add_str_to_docs(docs, str, key="new_meta"):
    fn = lambda x: [str] * len(x)
    return add_docs_metadata(docs, fn=fn, key=key)



from langchain_ray.imports import *
from langchain_ray.utils import *
from ray.util.queue import Queue, Empty
from langchain_ray.driver import redis_kv_store


def get_task_from_kv_store(task_id, kv_store):
    task = kv_store.get(task_id)
    if task is None:
        raise Exception(f"No task entry found for task_id: {task_id}.")
    if type(task) != dict:
        raise Exception(f"Wrong type for task with task_id: {task_id}.")
    if len(task) == 0:
        raise Exception(f"Empty dict for task_id: {task_id}.")
    return task


@ray.remote
def queue_consumer(id_, queue, redis_host="127.0.0.1", redis_port=6379) -> None:
    try:
        kv_store = redis_kv_store.KeyValueStore(redis_host=redis_host, redis_port=redis_port)
        while True:
            msg.info(f"Consumer {id_} going to block on queue.", spaced=True)
            task_data = queue.get(block=True)
            task_id = task_data["task_id"]
            tenant_id = task_data["tenant_id"]
            chain = task_data["chain"]
            chain_data = task_data["chain_data"]
            msg.info(f"Consumer {id_} got chain_data: {chain_data.dict()}.", spaced=True)
            try:
                task = get_task_from_kv_store(task_id, kv_store)
            except Exception as e:
                raise Exception(e)
            task["status"] = "TASK_STATUS_INPROGRESS"
            try:
                kv_store.insert(task_id, task)
                res = chain(chain_data)
                try:
                    task = get_task_from_kv_store(task_id, kv_store)
                except Exception as e:
                    raise Exception(e)
                task["status"] = "TASK_STATUS_FINISHED"
                task["results"] = json.dumps(res)
                msg.info(f"Inserting task_id: {task_id}.", spaced=True)
                kv_store.insert(task_id, task)
                msg.info(f"Inserted task_id: {task_id}.", spaced=True)
            except Exception as e:
                msg.fail(f"Queue Consumer failed with error: {e}", spaced=True)
                try:
                    task = get_task_from_kv_store(task_id, kv_store)
                except Exception as e:
                    raise Exception(e)
                task["status"] = "TASK_STATUS_ERROR"
                task["error"] = f"Queue Consumer failed with error: {e}."
                kv_store.insert(task_id, task)
    except Empty:
        pass


class Ingress:
    def __init__(
        self,
        max_queue_size=10,
        num_task_consumers=1,
        redis_host="127.0.0.1",
        redis_port=6379,
    ):
        try:
            self.kv_store = redis_kv_store.KeyValueStore(
                redis_host=redis_host, redis_port=redis_port
            )
            self.task_queue = Queue(
                maxsize=max_queue_size, actor_options={"num_cpus": 1, "num_gpus": 0}
            )
            self.task_consumers = []
            for i in range(num_task_consumers):
                self.task_consumers.append(
                    queue_consumer.remote(
                        id_=i,
                        queue=self.task_queue,
                        redis_host=redis_host,
                        redis_port=redis_port,
                    )
                )
        except Exception as e:
            msg.fail(f"Ingress init failed with error: {e}", spaced=True)

    async def action(self, data):
        if self.task_queue.full():
            raise Exception("\nTask Queue is Full. Please retry later.\n")
        chain_data = data["chain_data"]
        task_data = {"chain": data["chain"], "chain_data": chain_data}
        try:
            task_id = gen_random_string(16)
            tenant_id = chain_data.tenant_id
            self.kv_store.insert(task_id, {"status": "TASK_STATUS_CREATED"})
            task_data["task_id"] = task_id
            task_data["tenant_id"] = tenant_id
            self.task_queue.put(task_data, block=True)
            return {"task_id": task_id, "tenant_id": tenant_id}
        except Exception as e:
            raise Exception(f"Initiating Task failed with error: {e}")



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/indexing/00_indexing_utils.ipynb.

# %% auto 0
__all__ = ['index_names', 'docs_to_ems', 'add_ems_to_docs', 'docs_to_faiss', 'search_faiss']

# %% ../../nbs/indexing/00_indexing_utils.ipynb 2
from dreamai.imports import *
from ..imports import *
from ..utils import *
from ..chains import *

# %% ../../nbs/indexing/00_indexing_utils.ipynb 4
def index_names(index_folder, index_name):
    return [
        f.stem for f in Path(index_folder).glob("*.faiss") if f.stem.startswith(index_name)
    ]


def docs_to_ems(docs, ems_model):
    return ems_model.encode([doc.page_content for doc in docs]).tolist()


def add_ems_to_docs(docs, ems_model, key="embeddings"):
    fn = partial(docs_to_ems, ems_model=ems_model)
    return add_docs_metadata(docs, fn, key)


def docs_to_faiss(
    docs,
    ems_model,
    index_folder="/media/hamza/data2/faiss_data/saved_indexes/",
    index_name="index",
):
    "Create a `FAISS` database from `Documents`."
    db = FAISS.from_documents(flatten_list(docs), ems_model)
    # Get an unused index path.
    index_path = (Path(index_folder) / index_name).with_suffix(".faiss")
    if index_path.exists():
        index_path = find_alternate_path(index_path, first_idx=1, verbose=False)
    # Save the database to the index path.
    index_name = index_path.stem
    db.save_local(index_folder, index_name)
    return docs


def search_faiss(index_folder, index_names, query, ems_model, filter=None, k=2):
    # print(f"\n\nSearching {index_names} in {index_folder} for {query}!!!\n\n")
    if path_or_str(index_folder):
        index_folder = [index_folder]
    if path_or_str(index_names):
        index_names = [index_names]
    if path_or_str(query):
        query = [query]
    if is_iter(k):
        k = int(k[0])
    q_sims = []
    for index_folder, index_name, query in zip(index_folder, index_names, query):
        # print(f"\n\nSearching {index_name} in {index_folder} for {query}\n\n")
        db = FAISS.load_local(index_folder, embeddings=ems_model, index_name=index_name)
        q_sims.append(db.similarity_search_with_score(query, filter=filter, k=k))
    # print(f"\n\nQSIMS: {len([q_sims])}, INDEXES: {len(index_names)}\n\n")
    return [[q_sims]]



from dreamai.imports import *
from dreamai.core import *

import ray
import uuid
import torch
import demoji
import inspect
import subprocess
from time import time
from ray import serve
from ray import data as rd
from cleantext import clean
from setfit import SetFitModel
from fastai.torch_core import default_device
from sentence_transformers import SentenceTransformer
from transformers import (
    pipeline,
    AutoModel,
    AutoTokenizer,
    AutoModelForTokenClassification,
    AutoModelForSequenceClassification,
)

from langchain.schema import Document
from langchain.llms import HuggingFacePipeline
from langchain.vectorstores import Chroma, FAISS
from langchain.document_loaders import PyPDFLoader
from langchain.document_loaders.pdf import BasePDFLoader
from langchain.document_loaders.blob_loaders import Blob
from langchain.document_loaders.base import BaseBlobParser
from langchain.document_loaders.csv_loader import CSVLoader
from langchain.prompts.pipeline import PipelinePromptTemplate
from langchain.text_splitter import RecursiveCharacterTextSplitter
from langchain.chains import TransformChain, SimpleSequentialChain, SequentialChain
from langchain.embeddings.sentence_transformer import SentenceTransformerEmbeddings
from langchain import (
    PromptTemplate,
    FewShotPromptTemplate,
    HuggingFaceHub,
    LLMChain,
)



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/pdf/01_pdf_chains.ipynb.

# %% auto 0
__all__ = ['pdf_to_docs_chain', 'add_cats_to_docs_chain']

# %% ../../nbs/pdf/01_pdf_chains.ipynb 2
from ..imports import *
from ..chains import *
from ..utils import *
from .utils import *


# %% ../../nbs/pdf/01_pdf_chains.ipynb 4
def pdf_to_docs_chain(
    splitter=None,
    chunk_size=200,
    chunk_overlap=20,
    separators=["\n\n", "\n", "(?<=\. )", " ", ""],
    add_start_index=True,
    proc=True,
    input_variables=["path"],
    output_variables=["docs"],
    verbose=False,
):
    """Chain that returns a list of `Documents` extracted from a PDF path.
    The path can be a single PDF path or a list of paths or a directory path."""
    return transform_chain(
        pdf_to_docs,
        transform_kwargs=dict(
            splitter=splitter,
            chunk_size=chunk_size,
            chunk_overlap=chunk_overlap,
            separators=separators,
            add_start_index=add_start_index,
            proc=proc,
        ),
        vars_kwargs_mapping={input_variables[0]: "path"},
        input_variables=input_variables,
        output_variables=output_variables,
        verbose=verbose,
    )


def add_cats_to_docs_chain(
    cats_model,
    input_variables=["docs"],
    output_variables=["cat_docs"],
    verbose=False,
):
    "Chain that adds the categories to a list of `Documents` using `cats_model`."
    return transform_chain(
        add_cats_to_docs,
        transform_kwargs=dict(cats_model=cats_model),
        input_variables=input_variables,
        output_variables=output_variables,
        vars_kwargs_mapping={input_variables[0]: "docs"},
        verbose=verbose,
    )




from ray import serve
from fastapi import FastAPI
from dreamai.imports import *
from langchain_ray.utils import *
from langchain_ray.chains import *
from langchain_ray.imports import *
from pydantic import BaseModel, Field
from langchain_ray.ner.utils import *
from langchain_ray.pdf.chains import *
from langchain_ray.ner.chains import *
from langchain_ray.indexing.chains import *
from fastapi.responses import JSONResponse



from langchain_ray.api.imports import *
from langchain_ray.driver.ingress_driver import *


class TNetData(BaseModel):
    data_path: str = Field(
        title="The PDF path.",
        description="It can be a single file, list of files, directory or list of directories.",
        example="/home/hamza/dev/HF/langchain_ray/resumes_5",
    )
    ems_folder: str = Field(
        title="The folder to write the embeddings to.",
        description="It must be a directory.",
        regex=".*\/$",
        example="/media/hamza/data2/faiss_data/saved_ems/",
    )
    tenant_id: str = Field(
        title="The tenant id.",
        description="It must be a string.",
        example="ten_123",
        default="ten_123",
    )


class ResumesData(TNetData):
    cats_folder: str = Field(
        title="The folder to write the categories to.",
        description="It must be a directory.",
        regex=".*\/$",
        example="/media/hamza/data2/faiss_data/saved_cats/",
    )
    do_ner: bool = Field(
        title="Whether to do NER or not.",
        description="It must be a boolean.",
        default=False,
    )


class ResumeData(ResumesData):
    data_path: str = Field(
        title="The PDF path.",
        description="It must be a pdf file. Not a directory.",
        regex=".*\.pdf",
        example="/home/hamza/dev/HF/langchain_ray/resumes_5/0bedb223-262c-4388-9756-093dd7905428.pdf",
    )


app = FastAPI()


@serve.deployment(
    num_replicas=1,
    ray_actor_options=dict(num_cpus=1, num_gpus=0.1),
    health_check_period_s=10,
    health_check_timeout_s=30,
)
@serve.ingress(app)
class TNetIngress(Ingress):
    def __init__(
        self,
        model_name="HamzaFarhan/PDFSegs",
        block_size=50,
        num_cpus=4,
        num_gpus=0.2,
        max_queue_size=10,
        num_task_consumers=1,
        redis_host="127.0.0.1",
        redis_port=6379,
        verbose=True,
    ):
        self.ray_chain_args = dict(
            block_size=block_size, num_cpus=num_cpus, num_gpus=num_gpus, verbose=verbose
        )
        self.verbose = verbose
        device = default_device()
        self.ems_model = SentenceTransformer(model_name, device=device)
        self.cats_model = SetFitModel.from_pretrained("HamzaFarhan/PDFSegs").to(device)
        self.e_ner = load_edu_model(device=device),
        self.j_ner = load_job_model(device=device),
        super().__init__(
            max_queue_size=max_queue_size,
            num_task_consumers=num_task_consumers,
            redis_host=redis_host,
            redis_port=redis_port,
        )

    def res_chain(self, resumes_data: ResumesData):
        chains = []
        output_vars = []
        
        docs_chain = pdf_to_docs_chain(
            input_variables=["data_path"], output_variables=["docs"], verbose=self.verbose
        )
        chains.append(docs_chain)
        output_vars.append(["docs"])
        # tnet_chain = add_str_to_docs_chain(
        #     input_variables=output_vars[-1],
        #     output_variables=["tnet_docs"],
        #     verbose=self.verbose,
        # )
        # chains.append(tnet_chain)
        # output_vars.append(["tnet_docs"])
        cats_chain = add_cats_to_docs_chain(
            cats_model=self.cats_model,
            input_variables=output_vars[-1],
            output_variables=["cat_docs"],
            verbose=self.verbose,
        )
        cats_chain = ray_chain(cats_chain, **self.ray_chain_args)
        chains.append(cats_chain)
        output_vars.append(["cat_docs"])
        if resumes_data.do_ner:
            ner_chain = add_ners_to_docs_chain(
                e_ner=self.e_ner,
                j_ner=self.j_ner,
                input_variables=output_vars[-1],
                output_variables=["ner_docs"],
                verbose=self.verbose,
            )
            ner_chain = ray_chain(ner_chain, **self.ray_chain_args)
            chains.append(ner_chain)
            output_vars.append(["ner_docs"])
        
        json_chain = docs_to_json_chain(
            json_folder=resumes_data.cats_folder,
            input_variables=output_vars[-1],
            output_variables=["json_docs"],
            verbose=self.verbose,
        )
        chains.append(json_chain)
        output_vars.append(["json_docs"])
        ems_chain = add_ems_to_docs_chain(
            self.ems_model,
            input_variables=output_vars[-1],
            output_variables=["ems_docs"],
            verbose=self.verbose,
        )
        ems_chain = ray_chain(ems_chain, **self.ray_chain_args)
        chains.append(ems_chain)
        output_vars.append(["ems_docs"])
        json_chain2 = docs_to_json_chain(
            json_folder=resumes_data.ems_folder,
            input_variables=output_vars[-1],
            output_variables=["final_docs"],
            verbose=self.verbose,
        )
        chains.append(json_chain2)
        return SequentialChain(
            chains=chains,
            input_variables=["data_path"],
            output_variables=["final_docs"],
            verbose=self.verbose,
        )

    def jobs_chain(self, jobs_data: TNetData):
        chain1 = pdf_to_docs_chain(
            input_variables=["data_path"], output_variables=["docs"], verbose=self.verbose
        )
        chain2 = add_ems_to_docs_chain(
            self.ems_model,
            input_variables=["docs"],
            output_variables=["ems_docs"],
            verbose=self.verbose,
        )
        chain2 = ray_chain(chain2, **self.ray_chain_args)
        chain3 = docs_to_json_chain(
            jobs_data.ems_folder,
            with_content=False,
            input_variables=["ems_docs"],
            output_variables=["final_docs"],
            verbose=self.verbose,
        )
        return SequentialChain(
            chains=[chain1, chain2, chain3],
            input_variables=["data_path"],
            output_variables=["final_docs"],
            verbose=self.verbose,
        )

    @app.post("/batchembedding/resumes")
    async def resumes(self, resumes_data: ResumesData):
        try:
            chain = self.res_chain(resumes_data)
        except Exception as e:
            msg.fail("Failed to create Resumes Chain.", spaced=True)
            raise Exception(e)
        msg.info("********* CALLING ACTION *********", spaced=True)
        data_dict = dict(chain=chain, chain_data=resumes_data)
        res = await self.action(data=data_dict)
        # torch.cuda.empty_cache()
        msg.good(f"RETURNING RESULTS = {res}", spaced=True)
        return JSONResponse(content=res)

    @app.post("/batchembedding/resume")
    async def resumes(self, resumes_data: ResumesData):
        t1 = time()
        try:
            chain = self.res_chain(resumes_data)
        except Exception as e:
            msg.fail("Failed to create Resumes Chain.", spaced=True)
            raise Exception(e)
        try:
            chain(dict(data_path=resumes_data.data_path))
        except Exception as e:
            msg.fail("Failed to run Resumes Chain.", spaced=True)
            raise Exception(e)
        t2 = time()
        time_taken = t2 - t1
        msg.good(f"Time taken: {time_taken:.2f} seconds.", spaced=True)
        res = dict(
            data_path=resumes_data.data_path,
            cats_folder=resumes_data.cats_folder,
            ems_folder=resumes_data.ems_folder,
            time_taken=time_taken,
        )
        torch.cuda.empty_cache()
        msg.good(f"RETURNING RESULTS = {res}", spaced=True)
        return JSONResponse(content=res)

    @app.post("/batchembedding/jobs")
    async def jobs(self, jobs_data: TNetData):
        t1 = time()
        try:
            chain = self.jobs_chain(jobs_data)
        except Exception as e:
            msg.fail("Failed to create Jobs Chain.", spaced=True)
            raise Exception(e)
        try:
            chain(dict(data_path=jobs_data.data_path))
        except Exception as e:
            msg.fail("Failed to run Jobs Chain.", spaced=True)
            raise Exception(e)
        t2 = time()
        time_taken = t2 - t1
        msg.good(f"Time taken: {time_taken:.2f} seconds.", spaced=True)
        res = dict(
            data_path=jobs_data.data_path,
            ems_folder=jobs_data.ems_folder,
            time_taken=time_taken,
        )
        torch.cuda.empty_cache()
        msg.good(f"RETURNING RESULTS = {res}", spaced=True)
        return JSONResponse(content=res)
